from pandac.PandaModules import *

import math


class BobbleAI:
  def __init__(self,pos):
    self.destX = pos[0]
    self.destY = pos[1]
    self.toDie = False

    self.forceLimit = 50.0
    self.torqueLimit = 100.0
    self.targSpeed = 5.0
    self.targRot = math.pi


  def step(self,body):
    # The below code is basically stearing behaviour to get to a destination (x,y), with no attempt to avoid other stuff...

    # Calculate the delta we want...
    pos = body.getPosition()
    deltaX = self.destX - pos[0]
    deltaY = self.destY - pos[1]
    dist = math.sqrt(deltaX*deltaX + deltaY*deltaY)

    dt = max(1.0/50.0,globalClock.getDt()) # Only used for lookahead, hence max so a high framerate doesn't cause instability.

    # Get vector of current facing, projecting to x-y plane...
    try:
      curr = Quat(body.getQuaternion()).getForward()
      curr[2] = 0.0
      curr.normalize()

      # Calculate the rotation we want to acheive, and then apply torquq accordingly - we only bother if delta is large enough...
      if dist>0.05:
        # Take the angle between the facing vector and to destination vector...
        try:
          curAng = math.atan2(curr[1],curr[0])
          targAng = math.atan2(deltaY/dist,deltaX/dist)
          angDelta = targAng - curAng

          # Calculate the relevant angular velocity change to get there straight away, taking into account the current angular velocity...
          curAngVel = body.getAngularVel()[2]
          targAngVel = angDelta/dt
          targAngVel = max(min(targAngVel,self.targRot),-self.targRot)
          accNeeded = (targAngVel - curAngVel) / dt

          # Calculate the torque to acheive the stated goal...
          torqueLength = body.getMass().getInertialTensor()[2][2] * accNeeded
          torqueLength = min(max(torqueLength,-self.torqueLimit),self.torqueLimit)

          body.addTorque(Vec3(0.0,0.0,torqueLength))
        except:
          pass # Above can fail in various ways - will usually start working again once parameters nudge to something less singular.

      # Now do the linear stuff - need to travel dist...
      # Get the current speed and target distance to travel...
      lv = body.getLinearVel()
      curSpeed = lv[0]*curr[0] + lv[1]*curr[1]

      targDist = deltaX * curr[0] + deltaY * curr[1]
      targDist = max(targDist,0.0) # No walking backwards
      if targDist < 0.2: targDist = 0.0 # So it just stops if close enough.
      
      # Calculate the target velocity...
      targSpeed = targDist/dt
      targSpeed = min(targSpeed,self.targSpeed)
      targAcc = (targSpeed - curSpeed) / dt

      # Calculate the force...
      force = body.getMass().getMagnitude() * targAcc
      force = max(min(force,self.forceLimit),-self.forceLimit)

      body.addForce(Vec3(curr[0]*force,curr[1]*force,0.0))

    except:
      pass # Top catch a potential facing normalisation error if the bobble is on its backside.

    return not self.toDie


  def setStrength(self,lin,rot = 1.0):
    """Uses multipliers of the default amount."""
    self.forceLimit = 50.0*lin
    self.torqueLimit = 100.0*rot

  def setSpeed(self,lin,rot = 1.0):
    """Uses multipliers of the default amount."""
    self.targSpeed = 5.0*lin
    self.targRot = math.pi*rot

  def setDest(self,x,y):
    self.destX = x
    self.destY = y

  def die(self):
    if self.toDie==False:
      self.toDie = True
      del self.bobbles.bodyToAI[str(self.body)]

  def setCol(self,col):
    self.mesh.setColor(col[0],col[1],col[2])



class Bobbles:
  def __init__(self,ode,lighting):
    self.ode = ode
    
    # Load the bobble mesh...
    self.mesh = loader.loadModel('data/bobble')
    self.mesh.flattenStrong()

    # Create the new bobble node...
    self.bNode = render.attachNewNode('bNode')
    lighting.lightNode(self.bNode)

    # Converts bodys strings to their AI object...
    self.bodyToAI = dict()
    
    # We need a list of bobbles, so we can simulate them...
    self.bl = []

    def bobbles(task):
      if self.ode.active():
        blNew = []
        for bobble in self.bl:
          mesh,body,geomBot,geomTop,ai = bobble
        
          mesh.setPosQuat(render,body.getPosition(),Quat(body.getQuaternion()))
          if ai.step(body):
            blNew.append(bobble)
          else:
            mesh.removeNode()
            body.destroy()
            geomBot.destroy()
            geomTop.destroy()

        self.bl = blNew
      return task.cont

    taskMgr.add(bobbles,'bobbles')

  def addBobble(self,pos,h = 0.0,col = (1.0,1.0,1.0),Ai = BobbleAI):
    # Create the mesh...
    mesh = self.mesh.copyTo(self.bNode)
    mesh.setColor(col[0],col[1],col[2])
    mesh.setPos(pos)
    mesh.setH(h)

    # Create the body...
    body = OdeBody(self.ode.getWorld())
    body.setPosition(mesh.getPos(render))
    body.setQuaternion(mesh.getQuat(render))

    # Setup mass...
    density = 50.0
    
    bot = OdeMass()
    bot.setSphere(density,0.5)
    bot.translate(0.0,0.0,0.5)
    
    top = OdeMass()
    top.setSphere(density,0.25)
    top.translate(0.0,0.0,1.25)

    mass = OdeMass()
    mass.add(bot)
    mass.add(top)
    mass.translate(-mass.getCenter())
    body.setMass(mass)

    # Setup collision detection...
    geomBot = OdeSphereGeom(self.ode.getSpace(),0.5)
    geomBot.setBody(body)
    geomBot.setOffsetPosition(0.0,0.0,0.5)
    
    geomTop = OdeSphereGeom(self.ode.getSpace(),0.25)
    geomTop.setBody(body)
    geomTop.setOffsetPosition(0.0,0.0,1.25)

    # Create the AI object...
    ai = Ai(body.getPosition())
    ai.mesh = mesh
    ai.body = body
    ai.bobbles = self
    self.bodyToAI[str(body)] = ai

    # Add it to the list to be simulated...
    self.bl.append((mesh,body,geomBot,geomTop,ai))
    
    return ai
