from direct.showbase import DirectObject

import random
import math


class CamControl(object,DirectObject.DirectObject):
  def __init__(self,lowX,highX,lowY,highY,focalX = 0.0,focalY = 0.0,angle = math.pi/4.0):
    # We want to do our own control...
    base.disableMouse()

    # Store the cameras bounds...
    self.lowX = lowX
    self.highX = highX
    self.lowY = lowY
    self.highY = highY

    # Store the cameras current positoon - in terms of a focal point, angle and elevation - from this we calculate the cameras actual position...
    self.focalX = focalX
    self.focalY = focalY
    self.focalZ = 0.0
    self.angle = angle
    self.elevation = 32.0 # Height - we calculate the distance from focal using it.

    self.minElev = 4.0
    self.maxElev = 64.0

    self.calcCamera()

    # Keyboard control parameters...
    self.moveAmount = 6.0
    self.rotateAmount = 3.0
    self.zoomAmount = 24.0
    self.wheelZoomAmount = 3.0
    
    # Key press state...
    self.goForwards = False
    self.goBackwards = False
    self.goLeft = False
    self.goRight = False
    self.rotLeft = False
    self.rotRight = False
    self.zoomIn = False
    self.zoomOut = False

    # Code to turn key press state into action via a task...
    def cameraTask(task):
      self.elevation = min(max(self.elevation,self.minElev),self.maxElev)
      speed = self.moveAmount * math.sqrt(self.elevation)
      
      # Movement...
      if self.goForwards:
        self.focalX -= speed * math.cos(self.angle) * globalClock.getDt()
        self.focalY -= speed * math.sin(self.angle) * globalClock.getDt()
      if self.goBackwards:
        self.focalX += speed * math.cos(self.angle) * globalClock.getDt()
        self.focalY += speed * math.sin(self.angle) * globalClock.getDt()
      if self.goLeft:
        self.focalX += speed * math.sin(self.angle) * globalClock.getDt()
        self.focalY -= speed * math.cos(self.angle) * globalClock.getDt()
      if self.goRight:
        self.focalX -= speed * math.sin(self.angle) * globalClock.getDt()
        self.focalY += speed * math.cos(self.angle) * globalClock.getDt()

      # Rotation...
      if self.rotLeft:
        self.angle -= self.rotateAmount * globalClock.getDt()
      if self.rotRight:
        self.angle += self.rotateAmount * globalClock.getDt()

      # Zoom...
      if self.zoomIn:
        self.elevation += self.zoomAmount * globalClock.getDt()
      if self.zoomOut:
        self.elevation -= self.zoomAmount * globalClock.getDt()
      self.elevation = min(max(self.elevation,self.minElev),self.maxElev)

      self.calcCamera()
      return task.cont

    taskMgr.add(cameraTask,'Camera')
    
    # Accept all the keypresses we respond to - update the keypress state...
    self.accept('d',self.__setattr__,['goRight',True])
    self.accept('d-up',self.__setattr__,['goRight',False])
    self.accept('a',self.__setattr__,['goLeft',True])
    self.accept('a-up',self.__setattr__,['goLeft',False])
    self.accept('w',self.__setattr__,['goForwards',True])
    self.accept('w-up',self.__setattr__,['goForwards',False])
    self.accept('s',self.__setattr__,['goBackwards',True])
    self.accept('s-up',self.__setattr__,['goBackwards',False])

    self.accept('arrow_right',self.__setattr__,['goRight',True])
    self.accept('arrow_right-up',self.__setattr__,['goRight',False])
    self.accept('arrow_left',self.__setattr__,['goLeft',True])
    self.accept('arrow_left-up',self.__setattr__,['goLeft',False])
    self.accept('arrow_up',self.__setattr__,['goForwards',True])
    self.accept('arrow_up-up',self.__setattr__,['goForwards',False])
    self.accept('arrow_down',self.__setattr__,['goBackwards',True])
    self.accept('arrow_down-up',self.__setattr__,['goBackwards',False])

    self.accept('q',self.__setattr__,['rotLeft',True])
    self.accept('q-up',self.__setattr__,['rotLeft',False])
    self.accept('e',self.__setattr__,['rotRight',True])
    self.accept('e-up',self.__setattr__,['rotRight',False])
    self.accept('delete',self.__setattr__,['rotLeft',True])
    self.accept('delete-up',self.__setattr__,['rotLeft',False])
    self.accept('page_down',self.__setattr__,['rotRight',True])
    self.accept('page_down-up',self.__setattr__,['rotRight',False])

    self.accept('r',self.__setattr__,['zoomIn',True])
    self.accept('r-up',self.__setattr__,['zoomIn',False])
    self.accept('f',self.__setattr__,['zoomOut',True])
    self.accept('f-up',self.__setattr__,['zoomOut',False])
    self.accept('z',self.__setattr__,['zoomIn',True])
    self.accept('z-up',self.__setattr__,['zoomIn',False])
    self.accept('c',self.__setattr__,['zoomOut',True])
    self.accept('c-up',self.__setattr__,['zoomOut',False])
    self.accept('home',self.__setattr__,['zoomIn',True])
    self.accept('home-up',self.__setattr__,['zoomIn',False])
    self.accept('end',self.__setattr__,['zoomOut',True])
    self.accept('end-up',self.__setattr__,['zoomOut',False])

    self.accept('wheel_up',self.offsetElevation,[-self.wheelZoomAmount])
    self.accept('wheel_down',self.offsetElevation,[self.wheelZoomAmount])


  def calcCamera(self):
    dist = 0.5*self.maxElev*math.sin(math.pi*self.elevation/self.maxElev) + 0.1
    x = self.focalX + dist*math.cos(self.angle)
    y = self.focalY + dist*math.sin(self.angle)
    
    base.camera.setPos(x,y,self.focalZ + self.elevation)
    base.camera.lookAt(self.focalX,self.focalY,self.focalZ)

  def offsetElevation(self,amount):
    self.elevation += amount
